#include "runner.hpp"

#include <array>
#include <cerrno>
#include <poll.h>
#include <spawn.h>
#include <system_error>
#include <unistd.h>
#include <wait.h>

namespace sk {
runner::runner(runner_backend backend) : backend{backend} {}

run_result runner::run_blocking(const program &program) {
    int in_pipe[2];
    int out_pipe[2];
    int err_pipe[2];
    if (pipe(in_pipe) == -1 || pipe(out_pipe) == -1 || pipe(err_pipe) == -1) {
        throw std::system_error{errno, std::generic_category()};
    }

    posix_spawn_file_actions_t actions;
    posix_spawn_file_actions_init(&actions);
    posix_spawn_file_actions_addclose(&actions, in_pipe[1]);
    posix_spawn_file_actions_addclose(&actions, out_pipe[0]);
    posix_spawn_file_actions_addclose(&actions, err_pipe[0]);
    posix_spawn_file_actions_adddup2(&actions, in_pipe[0], STDIN_FILENO);
    posix_spawn_file_actions_adddup2(&actions, out_pipe[1], STDOUT_FILENO);
    posix_spawn_file_actions_adddup2(&actions, err_pipe[1], STDERR_FILENO);
    posix_spawn_file_actions_addclose(&actions, in_pipe[0]);
    posix_spawn_file_actions_addclose(&actions, out_pipe[1]);
    posix_spawn_file_actions_addclose(&actions, err_pipe[1]);
    const char *const docker_args[] = {"docker", "run", "--rm", "-i", "--pull=never", "--cap-drop=ALL", "--network=none", "--memory=64m", "--memory-swap=64m", "--pids-limit=128", program.image.c_str(), nullptr};
    const char *const bwrap_args[] = {"bwrap", "--ro-bind", "/usr", "/usr", "--dir", "/tmp", "--dir", "/var", "--proc", "/proc", "--dev", "/dev", "--symlink", "usr/lib", "/lib", "--symlink", "usr/lib64", "/lib64", "--symlink", "usr/bin", "/bin", "--symlink", "usr/sbin", "/sbin", "--unshare-all", "/bin/sh", nullptr};
    const char *const *args = docker_args;
    if (backend == runner_backend::BubbleWrap) {
        args = bwrap_args;
    }
    pid_t pid;
    int exit_code;
    if (posix_spawnp(&pid, args[0], &actions, nullptr, const_cast<char *const *>(args), nullptr) != 0) {
        throw std::system_error{errno, std::generic_category()};
    }

    close(in_pipe[0]);
    close(out_pipe[1]);
    close(err_pipe[1]);

    size_t len = program.code.size();
    size_t window = 0;
    const char *data = program.code.data();

    std::array<char, 1024> buffer{};
    std::string out;
    std::string err;
    std::array<pollfd, 3> plist = {pollfd{in_pipe[1], POLLOUT | POLLHUP, 0}, pollfd{out_pipe[0], POLLIN, 0}, pollfd{err_pipe[0], POLLIN, 0}};
    pollfd *pfds = plist.data();
    nfds_t nfds = plist.size();

    int res;
    while ((res = poll(pfds, nfds, -1)) > 0 || (res == -1 && errno == EINTR)) {
        if (res == -1) {
            // Interrupted by a signal, retry
            continue;
        }

        nfds_t polled_fds = 0;
        if (nfds == plist.size()) {
            // Poll stdin when we still have data to write
            if (pfds[0].revents & (POLLHUP | POLLERR)) {
                // Closing input prematurely
                close(in_pipe[1]);
                --nfds;
                ++pfds;
                ++polled_fds;
            } else if (pfds[0].revents & POLLOUT) {
                if (window >= len) {
                    // End input
                    close(in_pipe[1]);
                    --nfds;
                    ++pfds;
                } else {
                    // Write the current data window
                    size_t l = std::min(len - window, buffer.size());
                    write(pfds[0].fd, data + window, l);
                    window += l;
                }
                ++polled_fds;
            }
        }

        // Poll stdout and stderr
        for (nfds_t i = nfds - 2; i < nfds; ++i) {
            if (pfds[i].revents & POLLIN) {
                ssize_t bytes_read = read(pfds[i].fd, buffer.data(), buffer.size());
                if (bytes_read == -1) {
                    throw std::system_error{errno, std::generic_category()};
                }
                if (pfds[i].fd == out_pipe[0]) {
                    out.append(buffer.data(), bytes_read);
                } else {
                    err.append(buffer.data(), bytes_read);
                }
                ++polled_fds;
            }
        }

        // If nothing was polled, we're done
        if (polled_fds == 0) {
            break;
        }
    }

    waitpid(pid, &exit_code, 0);
    close(out_pipe[0]);
    close(err_pipe[0]);

    posix_spawn_file_actions_destroy(&actions);
    return run_result{out, err};
}
}
